--[[
Name: AutoBarSearch
Author: Toadkiller of Proudmoore
Website: http://www.wowace.com/
]]
-- Copyright 2007+ Toadkiller of Proudmoore.

-- http://www.wowace.com/wiki/AutoBar
-- http://www.curse-gaming.com/en/wow/addons-4430-1-autobar-toadkiller.html


local AceOO = AceLibrary("AceOO-2.0")
AutoBarSearch = {}

local searchSpace, items


-- Recycle lists will avoid garbage collection and memory thrashing but potentially grow over time
-- A simple 2 list aproach that recycles objects specific to that type of list so the bulk of operations should be only initing recycled objects.
-- We shall see if this is a good thing
local Recycle = AceOO.Class()
Recycle.virtual = true -- this means that it cannot be instantiated. (cannot call :new())
Recycle.prototype.recycleList = 0
Recycle.prototype.dataList = 0

function Recycle.prototype:init()
	Recycle.super.prototype.init(self) -- very important. Will fail without this.
	self.recycleList = {}
	self.dataList = {}
end

-- Returns a new or recycled list object
function Recycle.prototype:Create()
	if (self.recycleList[1]) then
		local i = table.maxn(self.recycleList)
		local x = self.recycleList[i]
		self.recycleList[i] = nil
		return x
	else
		return {}
	end
end

-- Adds some trash to the recycle list
-- do delete trash from the original list.
function Recycle.prototype:Destroy()
	table.insert(self.recycleList, trash)
end



-- The search space with all items to look for
-- Tracks client buttons that are looking (for proper deletion)
--		{ itemId = {buttonsIndex, ...} }
local SearchSpace = AceOO.Class(Recycle)

-- Add a list of itemIds for the given buttonsIndex
function SearchSpace.prototype:Add(itemId, buttonsIndex)
	local clientButtons = self.dataList[itemId]
	if (not clientButtons) then
		clientButtons = self:Create()
		self.dataList[itemId] = clientButtons
	end
	clientButtons[buttonsIndex] = true
end


-- Remove a list of itemIds for the given buttonsIndex
function SearchSpace.prototype:Delete(itemId, buttonsIndex)
	local clientButtons = self.dataList[itemId]
	if (clientButtons) then
		clientButtons[buttonsIndex] = nil
	end
	if (table.maxn(clientButtons) == 0) then
		self:Destroy(clientButtons)
		self.dataList[itemId] = nil
	end
end


-- Return the search space list.
-- Do not manipulate the list.  It is only for performance when checking if an itemId is searched for
function SearchSpace.prototype:GetList()
	return self.dataList
end



-- List of items to search for per slot.  No duplicates, highest priority overwrites.
-- Synced to SearchSpace.  All SearchSpace changes are via Items
-- Should only be changed via Config, or hunter pet food swithes
-- Priority is slotIndex.categoryIndex
--		{ itemId = {category, slotIndex, categoryIndex} }
local Items = AceOO.Class(Recycle)

function Items.prototype:init()
	Items.super.prototype.init(self) -- very important. Will fail without this.
	for buttonsIndex = 1, AUTOBAR_MAXBUTTONS, 1 do
		self.dataList[buttonsIndex] = {}
	end
end

-- Add a list of itemIds for the given buttonsIndex
function Items.prototype:Add(itemList, buttonsIndex, category, slotIndex)
--DEFAULT_CHAT_FRAME:AddMessage("Items.prototype:Add    category " .. tostring(category) .. " itemList " .. tostring(itemList))
	for i, itemId in pairs(itemList) do
		local buttonItems = self.dataList[buttonsIndex]
		local itemData = buttonItems[itemId]
		if (not itemData) then
			itemData = self:Create()
			buttonItems[itemId] = itemData
			itemData.category = category
			itemData.slotIndex = slotIndex
			itemData.categoryIndex = i
			AutoBarSearch.space:Add(itemId, buttonsIndex)
		elseif (slotIndex > itemData.slotIndex or (slotIndex == itemData.slotIndex and i >= itemData.categoryIndex)) then
			itemData.category = category
			itemData.slotIndex = slotIndex
			itemData.categoryIndex = i
		end
	end
end

-- Remove a list of itemIds for the given buttonsIndex
-- ToDo: on deletion reapply lower priority ones / track them from the start?
function Items.prototype:Delete(itemList, buttonsIndex, category, slotIndex)
	for i, itemId in pairs(itemList) do
		local buttonItems = self.dataList[buttonsIndex]
		local itemData = buttonItems[itemId]
		if (itemData and slotIndex == itemData.slotIndex and i == itemData.categoryIndex) then
			buttonItems[itemId] = nil
			self:Destroy(itemData)
			AutoBarSearch.space:Delete(itemId, buttonsIndex)
		end
	end
end

-- Populate the given button, or all if none specified
function Items.prototype:Populate(buttonsIndex)
	local function GetItems(category)
		if (type(category) == "string") then
			return category, AutoBar_Category_Info[category].items;
		elseif (type(category) == "table") then
			return nil, category;
		end
	end
	
	local startIndex = 1
	local endIndex = AUTOBAR_MAXBUTTONS
	if (buttonsIndex) then
		startIndex = buttonsIndex
		endIndex = buttonsIndex
	end
	local category, itemList
	for buttonsIndex = startIndex, endIndex, 1 do
		local buttonInfo = AutoBar.buttons[buttonsIndex]
		if (buttonInfo and buttonInfo[1]) then
			for slotIndex = 1, table.maxn(buttonInfo), 1 do
				category, itemList = GetItems(buttonInfo[slotIndex])
--DEFAULT_CHAT_FRAME:AddMessage("Items.prototype:Populate    itemList " .. tostring(itemList) .. " buttonsIndex " .. tostring(buttonsIndex) .. " category " .. tostring(category))
				if (itemList) then
					self:Add(itemList, buttonsIndex, category, slotIndex)
				end
			end
		end
	end
end

-- Return the buttons search list.
-- Do not manipulate the list.  It is only for performance when checking if an itemId is searched for
function Items.prototype:GetList(buttonsIndex)
	if (buttonsIndex) then
		return self.dataList[buttonsIndex]
	else
		return self.dataList
	end
end




-- Map of bag contents
-- Changes synced to Found
-- [bag][slot] = <itemId | nil>
local Bags = AceOO.Class()

function Bags.prototype:init()
	Bags.super.prototype.init(self) -- very important. Will fail without this.
	self.dataList = {}
	for bag = 0, 4, 1 do
		self.dataList[bag] = {}
	end
end

-- Add itemId to bag, slot
function Bags.prototype:Add(itemId, bag, slot)
	local slotList = self.dataList[bag]
	slotList[slot] = itemId
	AutoBarSearch.found:Add(itemId, bag, slot)
end

-- Remove itemId from bag, slot
function Bags.prototype:Delete(itemId, bag, slot)
	local slotList = self.dataList[bag]
	slotList[slot] = nil
	AutoBarSearch.found:Delete(itemId, bag, slot)
end

-- Scan the given bag.
function Bags.prototype:Scan(bag)
	local slotList = self.dataList[bag]
	local nSlots = GetContainerNumSlots(bag)
	local name, itemId, oldItemId

--DEFAULT_CHAT_FRAME:AddMessage("Bags.prototype:Scan bag " .. tostring(bag) .. " nSlots " .. tostring(nSlots) .. " slotList " .. tostring(slotList))

	-- ToDo: Clear out excess slots if bag got smaller

	for slot = 1, nSlots, 1 do
		name, itemId = AutoBar.LinkDecode(GetContainerItemLink(bag, slot))
		oldItemId = slotList[slot]

DEFAULT_CHAT_FRAME:AddMessage("Bags.prototype:Scan name " .. tostring(name) .. " itemId " .. tostring(itemId) .. " oldItemId " .. tostring(oldItemId))
		if (itemId) then
			if (oldItemId and oldItemId ~= itemId) then
				self:Delete(bag, slot)
				self:Add(itemId, bag, slot)
			elseif (not oldItemId) then
				self:Add(itemId, bag, slot)
			end
		elseif (not itemId and oldItemId) then
			self:Delete(oldItemId, bag, slot)
		end
	end
end

-- Scan the requested bags.
function Bags.prototype:ScanBags(bag0, bag1, bag2, bag3, bag4)
	for bag = 0, 4, 1 do
		if ("bag"..bag) then
			self:Scan(bag)
		end
	end
end

-- Return the bags list.
-- Do not manipulate the list.  It is only for performance.
function Bags.prototype:GetList()
	return self.dataList
end


-- Syncs to Bags and Current
-- itemId = { [bag, slot], ... }
local Found = AceOO.Class(Recycle)

function Found.prototype:init()
	Found.super.prototype.init(self) -- very important. Will fail without this.
	self.dataList = {}
end

-- Add itemId to bag, slot
function Found.prototype:Add(itemId, bag, slot)
	local itemData = self.dataList[itemId]
	if (not itemData) then
		itemData = self:Create()
		self.dataList[itemId] = itemData
		itemData[1] = bag
		itemData[2] = slot

		if (searchSpace[itemId]) then
			AutoBarSearch.current:Merge(itemId)
		end
--
--		-- ToDo: Hmmm, this should not be necessary
--		for i = table.maxn(itemData), 3, -1 do
--			itemData[i] = nil
--		end
	else
		local bFound = false
		local nItems = table.maxn(itemData)
		for i = 1, nItems - 1, 2 do
			if (itemData[i] and itemData[i] == bag and itemData[i + 1] == slot) then
				bFound = true
				break
			end
		end
		if (not found) then
			itemData[nItems + 1] = bag
			itemData[nItems + 2] = slot
		end
	end
end

-- Remove bag, slot for the itemId
function Found.prototype:Delete(itemId, bag, slot)
	local itemData = self.dataList[itemId]
	if (itemData) then
		local nItems = table.maxn(itemData)
		for i = 1, nItems - 1, 2 do
			if (itemData[i] and itemData[i] == bag and itemData[i + 1] == slot) then
				table.remove(itemData, i + 1);
				table.remove(itemData, i);
			end
		end
		if (not itemData[1]) then
			self.dataList[itemId] = nil
			self:Destroy(itemData)

			if (searchSpace[itemId]) then
				AutoBarSearch.current:Purge(itemId)
			end
		end
	end
end

-- Return the buttons found list.
-- Do not manipulate the list.  It is only for performance.
function Found.prototype:GetList()
	return self.dataList
end


-- list of found items for the button
-- bag, slot synced to Bags
-- { itemId, ... }
local Current = AceOO.Class()

function Current.prototype:init()
	Current.super.prototype.init(self) -- very important. Will fail without this.
	self.dataList = {}
	for buttonsIndex = 1, AUTOBAR_MAXBUTTONS, 1 do
		self.dataList[buttonsIndex] = {}
	end
end

-- Add the brand new item to any interrested buttons
function Current.prototype:Merge(itemId)
	for buttonsIndex = 1, AUTOBAR_MAXBUTTONS, 1 do
		local searchItems = items[buttonsIndex]
		if (searchItems and searchItems[itemId]) then
			local itemData = searchItems[itemId]
			self:Add(buttonsIndex, itemId)
			AutoBarSearch.sorted:Add(buttonsIndex, itemId, itemData.slotIndex, itemData.categoryIndex)
		end
	end
end

-- Purge the defunct item from its client buttons
function Current.prototype:Purge(itemId)
	for buttonsIndex = 1, AUTOBAR_MAXBUTTONS, 1 do
		local searchItems = items[buttonsIndex]
		if (searchItems and searchItems[itemId]) then
			self:Delete(buttonsIndex, itemId)
			AutoBarSearch.sorted:Delete(buttonsIndex, itemId)
		end
	end
end

-- Add the found item to the list of itemIds for the given buttonsIndex
function Current.prototype:Add(buttonsIndex, itemId)
	local buttonItems = self.dataList[buttonsIndex]
	buttonItems[itemId] = true
end

-- Remove the found item from the list of itemIds for the given buttonsIndex
-- ToDo: on deletion reapply lower priority ones / track them from the start?
function Current.prototype:Delete(buttonsIndex, itemId)
	local buttonItems = self.dataList[buttonsIndex]
	buttonItems[itemId] = nil
end

-- Return the buttons found list.
-- Do not manipulate the list.  It is only for performance.
function Current.prototype:GetList(buttonsIndex)
	if (buttonsIndex) then
		return self.dataList[buttonsIndex]
	else
		return self.dataList
	end
end



-- Sorted version of Current items for each button
-- Syncs to Items and Current
-- Verify / add items
-- n = { itemId, slotIndex, categoryIndex}, ... }
local Sorted = AceOO.Class(Recycle)

function Sorted.prototype:init()
	Sorted.super.prototype.init(self) -- very important. Will fail without this.
	for buttonsIndex = 1, AUTOBAR_MAXBUTTONS, 1 do
		self.dataList[buttonsIndex] = {}
	end
	self.dirtyList = {}
	self.dirty = false
end

-- Add the found item to the list of itemIds for the given buttonsIndex
function Sorted.prototype:Add(buttonsIndex, itemId, slotIndex, categoryIndex)
	local buttonItems = self.dataList[buttonsIndex]
	local bFound = false

	for i, itemData in ipairs(buttonItems) do
		if (itemData[1] == itemId) then
			bFound = true
			if (slotIndex > itemData[2]) then
				itemData[2] = slotIndex
				itemData[3] = categoryIndex
				-- Do not need a subcheck for categoryIndex, as itemId should be unique per category
			end
			break
		end
	end
	if (not bFound) then
		local itemData = self:Create()
		table.insert(buttonItems, itemData)
		itemData[1] = itemId
		itemData[2] = slotIndex
		itemData[3] = categoryIndex
	end
	self.dirtyList[buttonsIndex] = true
	self.dirty = true
end

-- Remove the found item from the list of itemIds for the given buttonsIndex
-- ToDo: on deletion reapply lower priority ones / track them from the start?
function Sorted.prototype:Delete(buttonsIndex, itemId)
	local buttonItems = self.dataList[buttonsIndex]
	local bFound = false

	for i, itemData in ipairs(buttonItems) do
		if (itemData[1] == itemId) then
			table.remove(buttonItems, i)
			self:Destroy(itemData)
			break
		end
	end
	self.dirtyList[buttonsIndex] = true
	self.dirty = true
end

local function SortBySlotCategory(a, b)
	if (a and b) then
		if (a[2] == b[2] and a[3] and b[3]) then
			return a[3] < b[3];
		else
			return a[2] < b[2];
		end
	else
		return true;
	end
end

-- Update any dirty lists.
function Sorted.prototype:Update()
	if (self.dirty) then
		for buttonsIndex = 1, AUTOBAR_MAXBUTTONS, 1 do
			if (self.dirtyList[buttonsIndex]) then
				local sortList = self.dataList[buttonsIndex]
				if (sortList) then
					table.sort(sortList, SortBySlotCategory);
				end
			end
		end
		self.dirty = false
	end
end

-- Return the buttons sorted list.
-- Do not manipulate the list.  It is only for performance.
function Sorted.prototype:GetList(buttonsIndex)
	if (buttonsIndex) then
		return self.dataList[buttonsIndex]
	else
		return self.dataList
	end
end



-- Only call this once
function AutoBarSearch:Initialize()
	AutoBarSearch.space = SearchSpace:new()		-- All items to search for
	AutoBarSearch.items = Items:new()			-- Items to search for for each button + category etc.

	AutoBarSearch.sorted = Sorted:new()			-- Sorted version of Current items

	AutoBarSearch.current = Current:new()		-- Current items found for each button (Found intersect Items)
	AutoBarSearch.found = Found:new()			-- All items found in Bags + list of bag, slot found in

	AutoBarSearch.bags = Bags:new()				-- Map of bag contents

	searchSpace = AutoBarSearch.space:GetList()
	items = AutoBarSearch.items:GetList()
end


-- Test harness		/script AutoBarSearch:Test()
function AutoBarSearch:Test()
	if (DevTools_Dump) then
--		AutoBarSearch.items:Add({4536}, 1, nil, 1)
--		AutoBarSearch.items:Add({4757}, 1, nil, 2)
--		AutoBarSearch.items:Add({6948}, 1, nil, 3)
--		AutoBarSearch.items:Add({4536, 6948, 4757}, 1, nil, 4)
--		DEFAULT_CHAT_FRAME:AddMessage("\n\n Items Added [1]")
--		DevTools_Dump(AutoBarSearch.items:GetList(1))
--
--		AutoBarSearch.items:Delete({4536}, 1, nil, 1)
--		AutoBarSearch.items:Delete({4757}, 1, nil, 2)
--		AutoBarSearch.items:Delete({6948}, 1, nil, 3)
--		AutoBarSearch.items:Delete({4536, 6948, 4757}, 1, nil, 4)
--		DEFAULT_CHAT_FRAME:AddMessage("\n\n Items Deleted [1]")
--		DevTools_Dump(AutoBarSearch.items:GetList(1))
--
--		DEFAULT_CHAT_FRAME:AddMessage("\n\n SearchSpace")
--		DevTools_Dump(AutoBarSearch.space:GetList())

		AutoBarSearch.items:Populate()
		DEFAULT_CHAT_FRAME:AddMessage("\n\n SearchSpace")
		DevTools_Dump(AutoBarSearch.space:GetList())

		AutoBarSearch.bags:ScanBags(bag0, bag1, bag2, bag3, bag4)

--		DEFAULT_CHAT_FRAME:AddMessage("\n\n Bags")
--		DevTools_Dump(AutoBarSearch.bags:GetList())
--
--		DEFAULT_CHAT_FRAME:AddMessage("\n\n Found")
--		DevTools_Dump(AutoBarSearch.found:GetList())
--
--		DEFAULT_CHAT_FRAME:AddMessage("\n\n Current")
--		DevTools_Dump(AutoBarSearch.current:GetList(1))

		AutoBarSearch.sorted:Update()
		DEFAULT_CHAT_FRAME:AddMessage("\n\n Sorted")
		DevTools_Dump(AutoBarSearch.sorted:GetList())
	end
end


